'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var React = require('react');
var index_cjs = require('../integration/index.cjs');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var React__default = /*#__PURE__*/_interopDefaultLegacy(React);

/**
 * React script component created for these scenarios:
 * - SSR/Static rendered HTML, without Client JS running
 * - Client JS only, without starting with rendered SSR/Static HTML
 * - SSR/Static rendered HTML, that is then hydrated with Client JS
 *
 * This component ensures each scenario works, and that the script is only
 * executed once (to include not re-executing during Hot Module Reloading).
 * Also ensures it does nothing for scripts with [type="text/partytown"].
 */
const PartytownScript = ({ id, innerHTML, type }) => {
    React.useEffect(() => {
        if (typeof document !== 'undefined' && type !== index_cjs.SCRIPT_TYPE) {
            if (!document.querySelector(`script[data-pt-script="${id}"]`)) {
                // Client side, that's not a Partytown script, that should run on main
                // SSR/Static script does not exist, this must be client only
                // append the script script that should have run and append to the head
                const scriptElm = document.createElement('script');
                scriptElm.dataset.ptScript = id;
                scriptElm.innerHTML = innerHTML;
                if (type) {
                    scriptElm.setAttribute('type', type);
                }
                document.head.appendChild(scriptElm);
            }
        }
    }, []);
    // `dangerouslySetInnerHTML` only works for scripts rendered as HTML from SSR/Static.
    // Added code will set the [type="data-pt-script"] attribute with the unique id.
    // If this code renders as SSR/Static HTML, then it'll execute and add the attribute
    // which will tell the Client JS of the component to NOT add the same script to the head.
    let jsxInnerHTML = innerHTML;
    if (type !== index_cjs.SCRIPT_TYPE) {
        jsxInnerHTML = `document.currentScript.dataset.ptScript=${JSON.stringify(id)};` + jsxInnerHTML;
    }
    return React__default["default"].createElement("script", { dangerouslySetInnerHTML: { __html: jsxInnerHTML }, type: type });
};

/**
 * The React `<Partytown/>` component should be placed within the `<head>`
 * of the document, and after any integrations such as `<GoogleTagManager/>`.
 *
 * @public
 */
const Partytown = (props) => (React__default["default"].createElement(PartytownScript, { id: "lib", innerHTML: index_cjs.partytownSnippet({ ...props }) }));

/**
 * https://developers.google.com/tag-manager/quickstart
 */
const googleTagManager = (containerId) => `
/* Google Tag Manager */
(function(w,d,s,l,i){w[l]=w[l]||[];w[l].push({'gtm.start':new Date().getTime(),event:'gtm.js'});var f=d.getElementsByTagName(s)[0],j=d.createElement(s),dl=l!='dataLayer'?'&l='+l:'';j.async=true;j.src='https://www.googletagmanager.com/gtm.js?id='+i+dl;f.parentNode.insertBefore(j,f);})(window,document,'script','dataLayer',${JSON.stringify(containerId)});
/* End Google Tag Manager */`;

/**
 * The Google Tag Manager Partytown component should be added after the opening `<head>`
 * tag, but before the `<Partytown/>` component. This component will add the
 * [Data Layer](https://developers.google.com/tag-manager/devguide) to the main thread
 * window, and will load GTM within the web worker. Any updates to `dataLayer.push(...)`
 * will be forwarded to the Partytown web worker.
 *
 * https://developers.google.com/tag-manager/quickstart
 *
 * @public
 */
const GoogleTagManager = ({ containerId }) => {
    return (React__default["default"].createElement(React.Fragment, null,
        React__default["default"].createElement(PartytownScript, { id: "gtm", innerHTML: index_cjs.appendForwardProperty('dataLayer', 1) }),
        React__default["default"].createElement(PartytownScript, { id: "gtm-pt", innerHTML: googleTagManager(containerId), type: index_cjs.SCRIPT_TYPE })));
};
/**
 * The GTM No Script component should be added immediately after the opening `<body>` tag.
 *
 * https://developers.google.com/tag-manager/quickstart
 *
 * @public
 */
const GoogleTagManagerNoScript = ({ containerId }) => (React__default["default"].createElement("noscript", null,
    React__default["default"].createElement("iframe", { src: 'https://www.googletagmanager.com/ns.html?id=' + encodeURIComponent(containerId), height: "0", width: "0", style: { display: 'none', visibility: 'hidden' } })));

exports.GoogleTagManager = GoogleTagManager;
exports.GoogleTagManagerNoScript = GoogleTagManagerNoScript;
exports.Partytown = Partytown;
